@@ -15,17 +15,18 @@ module Test.Tasty.Ingredients.ConsoleReporter
   -- ** Test failure statistics
   , Statistics(..)
   , printStatistics
+  , printStatisticsNoTime
   -- ** Outputting results
   , TestOutput(..)
   , buildTestOutput
   , foldTestOutput
   ) where
 
+import Prelude hiding (fail)
 import Control.Monad.State hiding (fail)
 import Control.Monad.Reader hiding (fail,reader)
 import Control.Concurrent.STM
 import Control.Exception
-import Control.Applicative
 import Test.Tasty.Core
 import Test.Tasty.Run
 import Test.Tasty.Ingredients
@@ -38,14 +39,18 @@ import qualified Data.IntMap as IntMap
 import Data.Char
 import Data.Maybe
 import Data.Monoid
-import Data.Proxy
-import Data.Tagged
 import Data.Typeable
-import Data.Foldable hiding (concatMap,elem,sequence_)
-import Options.Applicative
-import Prelude hiding (fail)  -- Silence AMP and FTP import warnings
+import Options.Applicative hiding (str)
 import System.IO
 import System.Console.ANSI
+#if !MIN_VERSION_base(4,8,0)
+import Data.Proxy
+import Data.Foldable hiding (concatMap,elem,sequence_)
+#endif
+#if MIN_VERSION_base(4,9,0)
+import Data.Semigroup (Semigroup)
+import qualified Data.Semigroup (Semigroup((<>)))
+#endif
 
 --------------------------------------------------
 -- TestOutput base definitions
@@ -55,16 +60,17 @@ import System.Console.ANSI
 -- printing. It lets us have several different printing modes (normal; print
 -- failures only; quiet).
 --
--- @since 0.11.3
+-- @since 0.12
 data TestOutput
   = PrintTest
+      {- test name         -} String
       {- print test name   -} (IO ())
       {- print test result -} (Result -> IO ())
-      -- ^ Action that prints the test name and an action that renders the
-      -- result of the action.
-  | PrintHeading (IO ()) TestOutput
-      -- ^ Action that prints the heading of a test group and the 'TestOutput'
-      -- for that test group.
+      -- ^ Name of a test, an action that prints the test name, and an action
+      -- that renders the result of the action.
+  | PrintHeading String (IO ()) TestOutput
+      -- ^ Name of a test group, an action that prints the heading of a test
+      -- group and the 'TestOutput' for that test group.
   | Skip -- ^ Inactive test (e.g. not matching the current pattern)
   | Seq TestOutput TestOutput -- ^ Two sets of 'TestOuput' on the same level
 
@@ -73,6 +79,10 @@ data TestOutput
 instance Monoid TestOutput where
   mempty = Skip
   mappend = Seq
+#if MIN_VERSION_base(4,9,0)
+instance Semigroup TestOutput where
+  (<>) = mappend
+#endif
 
 type Level = Int
 
@@ -118,7 +128,7 @@ buildTestOutput opts tree =
             (if resultSuccessful result then infoOk else infoFail) $
               printf "%s%s\n" (indent $ level + 1) (formatDesc (level+1) rDesc)
 
-      return $ PrintTest printTestName printTestResult
+      return $ PrintTest name printTestName printTestResult
 
     runGroup :: TestName -> Ap (Reader Level) TestOutput -> Ap (Reader Level) TestOutput
     runGroup name grp = Ap $ do
@@ -126,7 +136,7 @@ buildTestOutput opts tree =
       let
         printHeading = printf "%s%s\n" (indent level) name
         printBody = runReader (getApp grp) (level + 1)
-      return $ PrintHeading printHeading printBody
+      return $ PrintHeading name printHeading printBody
 
   in
     flip runReader 0 $ getApp $
@@ -139,14 +149,14 @@ buildTestOutput opts tree =
 
 -- | Fold function for the 'TestOutput' tree into a 'Monoid'.
 --
--- @since 0.11.3
+-- @since 0.12
 foldTestOutput
   :: Monoid b
-  => (IO () -> IO Result -> (Result -> IO ()) -> b)
+  => (String -> IO () -> IO Result -> (Result -> IO ()) -> b)
   -- ^ Eliminator for test cases. The @IO ()@ prints the testname. The
   -- @IO Result@ blocks until the test is finished, returning it's 'Result'.
   -- The @Result -> IO ()@ function prints the formatted output.
-  -> (IO () -> b -> b)
+  -> (String -> IO () -> b -> b)
   -- ^ Eliminator for test groups. The @IO ()@ prints the test group's name.
   -- The @b@ is the result of folding the test group.
   -> TestOutput -- ^ The @TestOutput@ being rendered.
@@ -154,7 +164,7 @@ foldTestOutput
   -> b
 foldTestOutput foldTest foldHeading outputTree smap =
   flip evalState 0 $ getApp $ go outputTree where
-  go (PrintTest printName printResult) = Ap $ do
+  go (PrintTest name printName printResult) = Ap $ do
     ix <- get
     put $! ix + 1
     let
@@ -162,9 +172,9 @@ foldTestOutput foldTest foldHeading outputTree smap =
         fromMaybe (error "internal error: index out of bounds") $
         IntMap.lookup ix smap
       readStatusVar = getResultFromTVar statusVar
-    return $ foldTest printName readStatusVar printResult
-  go (PrintHeading printName printBody) = Ap $
-    foldHeading printName <$> getApp (go printBody)
+    return $ foldTest name printName readStatusVar printResult
+  go (PrintHeading name printName printBody) = Ap $
+    foldHeading name printName <$> getApp (go printBody)
   go (Seq a b) = mappend (go a) (go b)
   go Skip = mempty
 
@@ -175,36 +185,36 @@ foldTestOutput foldTest foldHeading outputTree smap =
 --------------------------------------------------
 -- {{{
 consoleOutput :: (?colors :: Bool) => TestOutput -> StatusMap -> IO ()
-consoleOutput output smap =
-  getTraversal . fst $ foldTestOutput foldTest foldHeading output smap
+consoleOutput toutput smap =
+  getTraversal . fst $ foldTestOutput foldTest foldHeading toutput smap
   where
-    foldTest printName getResult printResult =
+    foldTest _name printName getResult printResult =
       ( Traversal $ do
-          printName
+          printName :: IO ()
           r <- getResult
           printResult r
       , Any True)
-    foldHeading printHeading (printBody, Any nonempty) =
+    foldHeading _name printHeading (printBody, Any nonempty) =
       ( Traversal $ do
-          when nonempty $ do printHeading; getTraversal printBody
+          when nonempty $ do printHeading :: IO (); getTraversal printBody
       , Any nonempty
       )
 
 consoleOutputHidingSuccesses :: (?colors :: Bool) => TestOutput -> StatusMap -> IO ()
-consoleOutputHidingSuccesses output smap =
-  void . getApp $ foldTestOutput foldTest foldHeading output smap
+consoleOutputHidingSuccesses toutput smap =
+  void . getApp $ foldTestOutput foldTest foldHeading toutput smap
   where
-    foldTest printName getResult printResult =
+    foldTest _name printName getResult printResult =
       Ap $ do
-          printName
+          printName :: IO ()
           r <- getResult
           if resultSuccessful r
             then do clearThisLine; return $ Any False
-            else do printResult r; return $ Any True
+            else do printResult r :: IO (); return $ Any True
 
-    foldHeading printHeading printBody =
+    foldHeading _name printHeading printBody =
       Ap $ do
-        printHeading
+        printHeading :: IO ()
         Any failed <- getApp printBody
         unless failed clearAboveLine
         return $ Any failed
@@ -213,11 +223,11 @@ consoleOutputHidingSuccesses output smap =
     clearThisLine = do clearLine; setCursorColumn 0
 
 streamOutputHidingSuccesses :: (?colors :: Bool) => TestOutput -> StatusMap -> IO ()
-streamOutputHidingSuccesses output smap =
+streamOutputHidingSuccesses toutput smap =
   void . flip evalStateT [] . getApp $
-    foldTestOutput foldTest foldHeading output smap
+    foldTestOutput foldTest foldHeading toutput smap
   where
-    foldTest printName getResult printResult =
+    foldTest _name printName getResult printResult =
       Ap $ do
           r <- liftIO $ getResult
           if resultSuccessful r
@@ -228,12 +238,12 @@ streamOutputHidingSuccesses output smap =
 
               liftIO $ do
                 sequence_ $ reverse stack
-                printName
-                printResult r
+                printName :: IO ()
+                printResult r :: IO ()
 
               return $ Any True
 
-    foldHeading printHeading printBody =
+    foldHeading _name printHeading printBody =
       Ap $ do
         modify (printHeading :)
         Any failed <- getApp printBody
@@ -265,12 +275,21 @@ data Statistics = Statistics
 instance Monoid Statistics where
   Statistics t1 f1 `mappend` Statistics t2 f2 = Statistics (t1 + t2) (f1 + f2)
   mempty = Statistics 0 0
+#if MIN_VERSION_base(4,9,0)
+instance Semigroup Statistics where
+  (<>) = mappend
+#endif
 
 computeStatistics :: StatusMap -> IO Statistics
 computeStatistics = getApp . foldMap (\var -> Ap $
   (\r -> Statistics 1 (if resultSuccessful r then 0 else 1))
     <$> getResultFromTVar var)
 
+reportStatistics :: (?colors :: Bool) => Statistics -> IO ()
+reportStatistics st = case statFailures st of
+    0 -> ok $ printf "All %d tests passed" (statTotal st)
+    fs -> fail $ printf "%d out of %d tests failed" fs (statTotal st)
+
 -- | @printStatistics@ reports test success/failure statistics and time it took
 -- to run. The 'Time' results is intended to be filled in by the 'TestReporter'
 -- callback. The @colors@ ImplicitParam controls whether coloured output is
@@ -280,13 +299,17 @@ computeStatistics = getApp . foldMap (\var -> Ap $
 printStatistics :: (?colors :: Bool) => Statistics -> Time -> IO ()
 printStatistics st time = do
   printf "\n"
-
+  reportStatistics st
   case statFailures st of
-    0 -> do
-      ok $ printf "All %d tests passed (%.2fs)\n" (statTotal st) time
+    0 -> ok $ printf " (%.2fs)\n" time
+    _ -> fail $ printf " (%.2fs)\n" time
 
-    fs -> do
-      fail $ printf "%d out of %d tests failed (%.2fs)\n" fs (statTotal st) time
+-- | @printStatisticsNoTime@ reports test success/failure statistics
+-- The @colors@ ImplicitParam controls whether coloured output is used.
+--
+-- @since 0.12
+printStatisticsNoTime :: (?colors :: Bool) => Statistics -> IO ()
+printStatisticsNoTime st = reportStatistics st >> printf "\n"
 
 -- | Wait until
 --
@@ -380,14 +403,14 @@ consoleTestReporter =
             ?colors = useColor whenColor isTerm
 
           let
-            output = buildTestOutput opts tree
+            toutput = buildTestOutput opts tree
 
           case () of { _
             | hideSuccesses && isTerm ->
-                consoleOutputHidingSuccesses output smap
+                consoleOutputHidingSuccesses toutput smap
             | hideSuccesses && not isTerm ->
-                streamOutputHidingSuccesses output smap
-            | otherwise -> consoleOutput output smap
+                streamOutputHidingSuccesses toutput smap
+            | otherwise -> consoleOutput toutput smap
           }
 
           return $ \time -> do
@@ -400,7 +423,7 @@ newtype Quiet = Quiet Bool
   deriving (Eq, Ord, Typeable)
 instance IsOption Quiet where
   defaultValue = Quiet False
-  parseValue = fmap Quiet . safeRead
+  parseValue = fmap Quiet . safeReadBool
   optionName = return "quiet"
   optionHelp = return "Do not produce any output; indicate success only by the exit code"
   optionCLParser = mkFlagCLParser (short 'q') (Quiet True)
@@ -410,7 +433,7 @@ newtype HideSuccesses = HideSuccesses Bool
   deriving (Eq, Ord, Typeable)
 instance IsOption HideSuccesses where
   defaultValue = HideSuccesses False
-  parseValue = fmap HideSuccesses . safeRead
+  parseValue = fmap HideSuccesses . safeReadBool
   optionName = return "hide-successes"
   optionHelp = return "Do not print tests that passed successfully"
   optionCLParser = mkFlagCLParser mempty (HideSuccesses True)
@@ -429,15 +452,16 @@ instance IsOption UseColor where
   defaultValue = Auto
   parseValue = parseUseColor
   optionName = return "color"
-  optionHelp = return "When to use colored output. Options are 'never', 'always' and 'auto' (default: 'auto')"
+  optionHelp = return "When to use colored output (default: 'auto')"
+  optionCLParser = mkOptionCLParser $ metavar "never|always|auto"
 
 -- | @useColor when isTerm@ decides if colors should be used,
 --   where @isTerm@ indicates whether @stdout@ is a terminal device.
 --
 --   @since 0.11.3
 useColor :: UseColor -> Bool -> Bool
-useColor when isTerm =
-  case when of
+useColor when_ isTerm =
+  case when_ of
     Never  -> False
     Always -> True
     Auto   -> isTerm
@@ -510,6 +534,10 @@ instance Ord a => Monoid (Maximum a) where
   Maximum a `mappend` Maximum b = Maximum (a `max` b)
   MinusInfinity `mappend` a = a
   a `mappend` MinusInfinity = a
+#if MIN_VERSION_base(4,9,0)
+instance Ord a => Semigroup (Maximum a) where
+  (<>) = mappend
+#endif
 
 -- | Compute the amount of space needed to align "OK"s and "FAIL"s
 computeAlignment :: OptionSet -> TestTree -> Int
