@@ -1,5 +1,5 @@
 -- | Ingredient for listing test names
-{-# LANGUAGE GeneralizedNewtypeDeriving, DeriveDataTypeable #-}
+{-# LANGUAGE CPP, GeneralizedNewtypeDeriving, DeriveDataTypeable #-}
 module Test.Tasty.Ingredients.ListTests
   ( ListTests(..)
   , testsNames
@@ -8,7 +8,6 @@ module Test.Tasty.Ingredients.ListTests
 
 import Data.Proxy
 import Data.Typeable
-import Data.Monoid
 import Options.Applicative
 
 import Test.Tasty.Core
@@ -21,7 +20,7 @@ newtype ListTests = ListTests Bool
   deriving (Eq, Ord, Typeable)
 instance IsOption ListTests where
   defaultValue = ListTests False
-  parseValue = fmap ListTests . safeRead
+  parseValue = fmap ListTests . safeReadBool
   optionName = return "list-tests"
   optionHelp = return "Do not run the tests; just print their names"
   optionCLParser = mkFlagCLParser (short 'l') (ListTests True)
